/*
 * A system pager implementation residing in a separate task.
 *
 * Copyright (C) 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/util/util.h>

#include <ipc/mem_ipc.h>
#include <mem/memory_utils.h>
#include <systypes/base.h>

#include <stdio.h>

#include "external_pager.h"



#define DEBUG 0



/* A simple system pager also acting as a region mapper. */

ExternalPager::ExternalPager(address_t start, address_t end)
: ExecPager(start, end), NotificationSupport()
{
}

/* Handle a general exception. */

long ExternalPager::exception(l4_exc_regs_t regs, l4_snd_fpage_t *region)
{
  (void) region;

  printf("ExternalPager::exception(...) -> pfa = %lx, pc = %lx\n", l4_utcb_exc_pfa(&regs), l4_utcb_exc_pc(&regs));

#if DEBUG
  while (1)
    l4_sleep_forever();
#endif

  return L4_EOK;
}

/* Handle a page fault using any configured regions. */

long ExternalPager::page_fault(l4_umword_t pfa, l4_umword_t pc, l4_snd_fpage_t *region)
{
  l4_umword_t addr = pfa & ~7UL, flags = pfa & 7;

#if DEBUG
  printf("page_fault(%lx, %lx) -> %lx (%lx) -> ", pfa, pc, addr, flags);
#endif

  /* Find the first region whose start address is beyond the fault address,
     testing if any immediately preceding region contains the fault address. */

  MappedRegions::iterator it = _regions.upper_bound(addr);

  if (it != _regions.begin())
    it--;
  else
  {
    printf("not mapped at %lx for pc %lx\n", addr, pc);
    return -L4_ENOMEM;
  }

  /* Obtain the region and test if it contains the fault address. */

  MappedRegion &r = it->second;

  if ((addr >= r.start) && (addr < r.start + r.size))
  {
    l4_addr_t page_addr = trunc(addr, L4_PAGESIZE);
    map_flags_t map_flags = map_flags_for_fault(flags);

    region->fpage = l4_fpage(r.ds_start + (page_addr - r.start), L4_PAGESHIFT, map_flags & r.flags);
    region->snd_base = page_addr;

#if DEBUG
    printf("%lx...%lx from %lx...%lx offset %lx size %d rights %x ds %lx\n",
           r.start, region->snd_base,
           r.ds_start, l4_fpage_memaddr(region->fpage),
           addr - r.start,
           l4_fpage_size(region->fpage),
           l4_fpage_rights(region->fpage),
           r.ds);

    printf("%lx -> ", addr);

    for (unsigned int i = 0; i < sizeof(l4_umword_t); i++)
      printf("%02x", *((unsigned char *)(r.ds_start + (addr - r.start) + i)));

    printf("\n");
#endif

    if (r.flags & L4RE_RM_F_W)
      l4_touch_rw((const void *) (r.ds_start + (page_addr - r.start)), L4_PAGESIZE);
    else
      l4_touch_ro((const void *) (r.ds_start + (page_addr - r.start)), L4_PAGESIZE);

    return L4_EOK;
  }

#if DEBUG
  printf("not mapped!\n");
#endif

  return -L4_ENOMEM;
}

/* Attach a region for provision when page faults occur. This is required in
   the initialisation of a program by the C library which requires a region
   mapper. */

long ExternalPager::attach(address_t *start, address_t size, map_flags_t flags,
                           l4_cap_idx_t ds, address_t offset,
                           unsigned char align)
{
  long err = ExecPager::find(start, &size, flags, offset, align);

  if (!err)
  {
    /* Attach the provided dataspace.

       This is only done in this implementation to support the paging
       mechanism. In a region mapper residing within the actual task, the
       dataspace's map operation would be invoked to obtain mappings. */

    l4_addr_t ds_start;
    err = ipc_attach_dataspace(ds, size, (void **) &ds_start);

    if (err)
      return err;

    l4_touch_rw((const void *) ds_start, size);

    MappedRegion r(*start, size, flags & L4RE_DS_F_RIGHTS_MASK, ds, ds_start);
    add(r);
  }

  return err;
}

/* Receive signals from a task. */

long ExternalPager::signal(unsigned long sig, unsigned long val)
{
  notify_all(NOTIFY_TASK_SIGNAL, (notify_values_t) {sig, val});

  return L4_EOK;
}

/* Subscribe to notifications. */

long ExternalPager::subscribe(l4_cap_idx_t notifier, notify_flags_t flags)
{
  _notifier = notifier;
  return NotificationSupport::subscribe(notifier, flags);
}

/* Unsubscribe from notifications. */

long ExternalPager::unsubscribe()
{
  NotificationSupport::unsubscribe(_notifier);
  return L4_EOK;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
